#!/usr/bin/env python

import time
import sys
from operator import itemgetter

def fileLen(fname):
    with open(fname) as f:
        for i, l in enumerate(f):
            pass
    return i + 1

def init(buffer):
   from glob import globs, setGlobals
   from array import array

   # create spatial variables of specified number
   temp = buffer.pop(0)
   Vars = int(temp.split()[0].strip())+2
   TypeId = temp.split()[1].strip()
   
   # initialize constraints
   N = fileLen(globs['calc']+"/"+globs['calc']+".relations")
   setGlobals("size", N)
   Id = 0
   from bitcoding import B_dict
   with open(globs['calc']+"/"+globs['calc']+".identity") as f:
      Id = B_dict[f.readline().strip()]
  
   ConMatrix = [None for v in xrange(Vars)]
   for v in xrange(Vars):
      j = {}   
      j[v] = Id
      ConMatrix[v] = j

   # parse spatial CSP
   from parsecsp import parsecsp
   parsecsp(ConMatrix, buffer)
   
   return TypeId, ConMatrix

class Usage(Exception):
    def __init__(self, msg):
        self.msg = msg

# main function
def main(argv=None):

   if argv is None:
        argv = sys.argv

   # set some sys useful stuff
   sys.setrecursionlimit(1<<30)
   sys.setcheckinterval(10000)

   from optparse import OptionParser

   # initialize parser for command line arguments
   parser = OptionParser()
   
   # set parsing options
   parser.add_option("-c", "--calculus", type="string", dest="calc", default="rcc8", help="interaction mode: rcc8, or allen [default: %default]")
   parser.add_option("-d", "--print", action="store_true", dest="printsol", help="print solution to output")
   parser.add_option("-f", "--file", action="store", type="string", dest="filename")

   # parse command line arguments
   options, args = parser.parse_args()

   from glob import setGlobals
   #import argparse

   try:   
      # check and set globals for calculus information
      if not (options.calc == "rcc8" or options.calc == "allen"):
         raise Usage("Invalid option: " + options.calc)
      else:
         setGlobals('calc',options.calc)

   except Usage, err:
        print err.msg
        parser.print_help()
        return 1

   with open(options.filename) as f:
      total = 0
      meter = 0
      edges_in_complete_graph = 0
      total_edges = 0
      good_edges = 0
      buffer = []

      for i in f:
         buffer.append(i)

         if i.strip() == '.':
            start = elapsed = 0
            meter += 1

            from counters import counter

            TypeId, ConMatrix = init(buffer)
            buffer = []
            
            from triangulation import MCS, FIC
            from bitcoding import B_dict
               
            edges = set([])

            neighbors = tuple([set([]) for i in xrange(len(ConMatrix))])
            for n, i in enumerate(ConMatrix):
               for j in i:
                  if j > n:
                     edges.add((n,j))
                     neighbors[n].add(j)
                     neighbors[j].add(n)
            
            edges_in_complete_graph = len(ConMatrix)*((len(ConMatrix)-1)/2.0)
            total_edges += len(edges)

            fill = FIC(ConMatrix, neighbors, MCS(ConMatrix, neighbors))

            from bitcoding import B_dict
            DALL = B_dict['DALL']

            for i,j in fill:
               if (i,j) not in edges:
                  neighbors[i].add(j)
                  neighbors[j].add(i)

            for i,j in fill:
               if (i,j) not in edges:
                  ConMatrix[i][j] = DALL
                  ConMatrix[j][i] = DALL
                  
            from pc import PC as pc

            start = time.clock()
            solution = pc(ConMatrix,neighbors)         

            from bon import bon
            bonnes = len(bon(ConMatrix,neighbors))
            good_edges += bonnes
            elapsed = time.clock() - start
            total += elapsed

            # print values of complexity counters
            print '--------------------------------------'
            print 'Type ID: ', TypeId
            print '--------------------------------------'
            print 'Visited arcs: ', counter['arcs']
            print 'Checked constraints: ', counter['cons']
            print '--------------------------------------'
            print '# of relations in the complete graph: ', edges_in_complete_graph
            print '# of initial relations: ', len(edges)
            print '# of non-redundant relations: ', bonnes
            print '--------------------------------------'
            print 'Time: ', elapsed, 'sec'
            print '--------------------------------------'

            # print solution  
            if solution != False:
               print 'Path Consistent'
               if options.printsol == True:
                  from printm import printMatrix
                  printMatrix(ConMatrix)
            else:
               print 'Inconsistent'
            print '--------------------------------------'

            edges_in_complete_graph = 0
            counter['arcs'] = 0
            counter['cons'] = 0
            ConMatrix = None

      print '======================================'
      print 'Total time:', total, 'sec'
      print 'Avg time per CSP:', total/(1.0*meter), 'sec'
      print 'Total # of initial relations: ', total_edges
      print 'Total # of non-redundant relations: ', good_edges
      print 'Reduction (%): ', (good_edges-total_edges)/float(total_edges)*100.0
      print '======================================' 

if __name__ == '__main__':
   sys.exit(main())

