from comptab import fcomp
from bitcoding import B_dict
from inverse import inv
from collections import deque
from counters import counter
from itertools import combinations

DALL = B_dict['DALL']

# path consistency using van Beek weights
def PC(ConMatrix, neighbors):
   pq = deque([])                         # list of entries arranged in a heap
   entry_finder = set([])               # mapping of tasks to entries
   Vars = len(ConMatrix)

   def add_task(task):
       'Add a new task or update the priority of an existing task'
       if task not in entry_finder:
          entry_finder.add(task)
          pq.append(task)
          
   def pop_task():
       'Remove and return the lowest priority task. Raise KeyError if empty.'
       task = pq.popleft()
       entry_finder.discard(task)
       return task

   # initialize queue
   map(add_task,[(n,j) for n, i in enumerate(ConMatrix) for j in i if j > n and ConMatrix[n][j] != DALL])

   # as long as the queue is not empty, process it
   while pq:

      (i,j) = pop_task() # grab the appropriate relation
      
      counter['arcs'] += 1 # increment visited arcs counter 
  
      # create all triplets to be checked for path consistency
      for k in neighbors[i] & neighbors[j]:

          temp = fcomp[ConMatrix[k][i]-1][ConMatrix[i][j]-1]
          if temp != DALL:

             counter['cons'] += 1 # increment checked constraints counter
  
             # constrain arc (k,i,j)
             temp = temp & ConMatrix[k][j]
             if temp != ConMatrix[k][j]:
                if not temp: 
                   return False # inconsistency
                r = ConMatrix[k][j]
                ConMatrix[k][j] = temp
                ConMatrix[j][k] = inv[temp-1]
                if k < j:
                   add_task((k, j)) 
                else:
                   add_task((j, k))

          temp = fcomp[ConMatrix[i][j]-1][ConMatrix[j][k]-1]
          if temp != DALL:

             counter['cons'] += 1 # increment checked constraints counter
          
             # constrain arc (i,j,k) 
             temp = temp & ConMatrix[i][k] 
             if temp != ConMatrix[i][k]:
                if not temp: 
                   return False # inconsistency
                r = ConMatrix[i][k]
                ConMatrix[i][k] = temp
                ConMatrix[k][i] = inv[temp-1]
                if i < k:
                   add_task((i, k)) 
                else:
                   add_task((k, i))

   # the network is concistent and can't be refined further
   return True    
